"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
require("./types/fstream");
const fs_extra_1 = require("fs-extra");
const fstream_1 = require("fstream");
const tar_1 = require("tar");
const tar_stream_1 = __importDefault(require("tar-stream"));
const https_1 = __importStar(require("https"));
const zlib_1 = __importDefault(require("zlib"));
const http_1 = require("http");
const stream_1 = require("stream");
const string_decoder_1 = require("string_decoder");
const miscellany_1 = require("./miscellany");
const fs_1 = require("./fs");
const assertions_1 = require("./assertions");
async function readFile(path) {
    const res = await fs_extra_1.readFile(path, { encoding: "utf8" });
    if (res.includes("�")) {
        throw new Error(`Bad character in ${path}`);
    }
    return res;
}
exports.readFile = readFile;
function readFileSync(path) {
    const res = fs_extra_1.readFileSync(path, { encoding: "utf8" });
    if (res.includes("�")) {
        throw new Error(`Bad character in ${path}`);
    }
    return res;
}
exports.readFileSync = readFileSync;
/** If a file doesn't exist, warn and tell the step it should have been generated by. */
async function readFileAndWarn(generatedBy, filePath) {
    try {
        return await readJson(filePath);
    }
    catch (e) {
        console.error(`Run ${generatedBy} first!`);
        throw e;
    }
}
exports.readFileAndWarn = readFileAndWarn;
function readJsonSync(path) {
    return miscellany_1.parseJson(readFileSync(path));
}
exports.readJsonSync = readJsonSync;
async function readJson(path) {
    return miscellany_1.parseJson(await readFile(path));
}
exports.readJson = readJson;
function writeFile(path, content) {
    return fs_extra_1.writeFile(path, content, { encoding: "utf8" });
}
exports.writeFile = writeFile;
function writeJson(path, content, formatted = true) {
    return fs_extra_1.writeJson(path, content, { spaces: formatted ? 4 : 0 });
}
exports.writeJson = writeJson;
function streamOfString(text) {
    const s = new stream_1.Readable();
    s.push(text);
    s.push(null); // tslint:disable-line no-null-keyword
    return s;
}
exports.streamOfString = streamOfString;
function stringOfStream(stream, description) {
    const decoder = new string_decoder_1.StringDecoder("utf8");
    let body = "";
    stream.on("data", (data) => {
        body += decoder.write(data);
    });
    return new Promise((resolve, reject) => {
        stream.on("error", reject);
        stream.on("end", () => {
            body += decoder.end();
            if (body.includes("�")) {
                reject(`Bad character decode in ${description}`);
            }
            else {
                resolve(body);
            }
        });
    });
}
exports.stringOfStream = stringOfStream;
function streamDone(stream) {
    return new Promise((resolve, reject) => {
        stream.on("error", reject).on("finish", resolve);
    });
}
exports.streamDone = streamDone;
class Fetcher {
    constructor() {
        this.agent = new https_1.Agent({ keepAlive: true });
    }
    async fetchJson(options) {
        const text = await this.fetch(options);
        try {
            return JSON.parse(text);
        }
        catch (e) {
            throw new Error(`Bad response from server:\noptions: ${JSON.stringify(options)}\n\n${text}`);
        }
    }
    async fetch(options) {
        const maxRetries = options.retries === false || options.retries === undefined ? 0 : options.retries === true ? 10 : options.retries;
        for (let retries = maxRetries; retries > 1; retries--) {
            try {
                return await doRequest(options, https_1.request, this.agent);
            }
            catch (err) {
                if (!/EAI_AGAIN|ETIMEDOUT|ECONNRESET/.test(err.message)) {
                    throw err;
                }
            }
            await miscellany_1.sleep(1);
        }
        return doRequest(options, https_1.request, this.agent);
    }
}
exports.Fetcher = Fetcher;
/** Only used for testing. */
function makeHttpRequest(options) {
    return doRequest(options, http_1.request);
}
exports.makeHttpRequest = makeHttpRequest;
function doRequest(options, makeRequest, agent) {
    return new Promise((resolve, reject) => {
        const req = makeRequest({
            hostname: options.hostname,
            port: options.port,
            path: `/${options.path}`,
            agent,
            method: options.method || "GET",
            headers: options.headers
        }, res => {
            let text = "";
            res.on("data", (d) => {
                text += d;
            });
            res.on("error", reject);
            res.on("end", () => {
                resolve(text);
            });
        });
        if (options.body !== undefined) {
            req.write(options.body);
        }
        req.end();
    });
}
async function isDirectory(path) {
    return (await fs_extra_1.stat(path)).isDirectory();
}
exports.isDirectory = isDirectory;
exports.npmInstallFlags = "--ignore-scripts --no-shrinkwrap --no-package-lock --no-bin-links --no-save";
function downloadAndExtractFile(url) {
    return new Promise((resolve, reject) => {
        const root = new fs_1.Dir(undefined);
        function insertFile(path, content) {
            const components = path.split("/");
            const baseName = assertions_1.assertDefined(components.pop());
            let dir = root;
            for (const component of components) {
                dir = dir.subdir(component);
            }
            dir.set(baseName, content);
        }
        https_1.default
            .get(url, { timeout: 1000000 }, response => {
            const extract = tar_stream_1.default.extract();
            response.pipe(zlib_1.default.createGunzip()).pipe(extract);
            extract.on("entry", (header, stream, next) => {
                const name = assertions_1.assertDefined(miscellany_1.withoutStart(header.name, "DefinitelyTyped-master/"));
                switch (header.type) {
                    case "file":
                        stringOfStream(stream, name)
                            .then(s => {
                            insertFile(name, s);
                            next();
                        })
                            .catch(reject);
                        break;
                    case "directory":
                        next();
                        break;
                    default:
                        throw new Error(`Unexpected file system entry kind ${header.type}`);
                }
            });
            extract.on("error", reject);
            extract.on("finish", () => {
                resolve(new fs_1.InMemoryFS(root.finish(), ""));
            });
        })
            .on("error", reject);
    });
}
exports.downloadAndExtractFile = downloadAndExtractFile;
function gzip(input) {
    return input.pipe(zlib_1.default.createGzip());
}
exports.gzip = gzip;
function unGzip(input) {
    const output = zlib_1.default.createGunzip();
    input.pipe(output);
    return output;
}
exports.unGzip = unGzip;
function writeTgz(inputDirectory, outFileName) {
    return new Promise((resolve, reject) => {
        resolve(streamDone(createTgz(inputDirectory, reject).pipe(fs_extra_1.createWriteStream(outFileName))));
    });
}
exports.writeTgz = writeTgz;
// To output this for testing:
// `require("./dist/io").createTgz("./src", err => { throw err }).pipe(fs.createWriteStream("foo.tgz"))`
function createTgz(dir, onError) {
    return gzip(createTar(dir, onError));
}
exports.createTgz = createTgz;
function createTar(dir, onError) {
    const packer = tar_1.Pack({ noProprietary: true, path: dir }).on("error", onError);
    return fstream_1.Reader({ path: dir, type: "Directory", filter: addDirectoryExecutablePermission })
        .on("error", onError)
        .pipe(packer);
}
/**
 * Work around a bug where directories bundled on Windows do not have executable permission when extracted on Linux.
 * https://github.com/npm/node-tar/issues/7#issuecomment-17572926
 */
function addDirectoryExecutablePermission(entry) {
    if (entry.props.type === "Directory") {
        entry.props.mode = addExecutePermissionsFromReadPermissions(entry.props.mode);
    }
    return true;
}
function addExecutePermissionsFromReadPermissions(mode) {
    // Constant that gives execute permissions to owner, group, and others. "+x"
    const allExecutePermissions = 0o111;
    // Moves the bits for read permissions into the place for execute permissions.
    // In other words, a component will have execute permissions if it has read permissions.
    const readPermissionsAsExecutePermissions = (mode >>> 2) & allExecutePermissions; // tslint:disable-line no-bitwise
    // Add these additional execute permissions to the mode.
    return mode | readPermissionsAsExecutePermissions; // tslint:disable-line no-bitwise
}
//# sourceMappingURL=io.js.map